"""From: https://github.com/langchain-ai/langchain/blob/master/libs/langchain/langchain/output_parsers/json.py
"""
import json
import re
from typing import Any, Callable, Dict


class OutputParserError(ValueError):
    """输出解析异常"""

    pass


# Adapted from https://github.com/KillianLucas/open-interpreter/blob/main/interpreter/utils/parse_partial_json.py
# MIT License
def parse_partial_json(s: str, *, strict: bool = False) -> Any:
    """Parse a JSON string that may be missing closing braces.

    Args:
        s: The JSON string to parse.
        strict: Whether to use strict parsing. Defaults to False.

    Returns:
        The parsed JSON object as a Python dictionary.
    """
    # Attempt to parse the string as-is.
    try:
        return json.loads(s, strict=strict)
    except json.JSONDecodeError:
        pass

    # Initialize variables.
    new_s = ""
    stack = []
    is_inside_string = False
    escaped = False

    # Process each character in the string one at a time.
    for char in s:
        if is_inside_string:
            if char == '"' and not escaped:
                is_inside_string = False
            elif char == "\n" and not escaped:
                char = "\\n"  # Replace the newline character with the escape sequence. # noqa: PLW2901
            elif char == "\\":
                escaped = not escaped
            else:
                escaped = False
        elif char == '"':
            is_inside_string = True
            escaped = False
        elif char == "{":
            stack.append("}")
        elif char == "[":
            stack.append("]")
        elif char in ("}", "]"):
            if stack and stack[-1] == char:
                stack.pop()
            else:
                # Mismatched closing character; the input is malformed.
                return None

        # Append the processed character to the new string.
        new_s += char

    # If we're still inside a string at the end of processing,
    # we need to close the string.
    if is_inside_string:
        new_s += '"'

    # Close any remaining open structures in the reverse order that they were opened.
    for closing_char in reversed(stack):
        new_s += closing_char

    # Attempt to parse the modified string as JSON.
    try:
        return json.loads(new_s, strict=strict)
    except json.JSONDecodeError:
        # If we still can't parse the string as JSON, return None to indicate failure.
        return None


def _replace_new_line(match: re.Match[str]) -> str:
    value = match.group(2)
    value = re.sub(r"\n", r"\\n", value)
    value = re.sub(r"\r", r"\\r", value)
    value = re.sub(r"\t", r"\\t", value)
    value = re.sub(r'(?<!\\)"', r"\"", value)

    return match.group(1) + value + match.group(3)


def _custom_parser(multiline_string: str) -> str:
    """
    The LLM response for `action_input` may be a multiline
    string containing unescaped newlines, tabs or quotes. This function
    replaces those characters with their escaped counterparts.
    (newlines in JSON must be double-escaped: `\\n`)
    """
    if isinstance(multiline_string, (bytes, bytearray)):
        multiline_string = multiline_string.decode()

    multiline_string = re.sub(
        r'("action_input"\:\s*")(.*)(")',
        _replace_new_line,
        multiline_string,
        flags=re.DOTALL,
    )

    return multiline_string


def parse_json_markdown(json_string: str, *, parser: Callable[[str], Any] = json.loads) -> Any:
    """
    Parse a JSON string from a Markdown string.

    Args:
        json_string: The Markdown string.

    Returns:
        The parsed JSON object as a Python dictionary.
    """
    # Try to find JSON string within triple backticks
    match = re.search(r"```(json)?(.*)```", json_string, re.DOTALL)

    # If no match found, assume the entire string is a JSON string
    if match is None:
        json_str = json_string
    else:
        # If match found, use the content within the backticks
        json_str = match.group(2)

    # Strip whitespace and newlines from the start and end
    json_str = json_str.strip()

    # handle newlines and other special characters inside the returned value
    json_str = _custom_parser(json_str)

    # Parse the JSON string into a Python dictionary
    parsed = parser(json_str)

    return parsed


class SimpleJsonOutputParser:
    """Parse the output of an LLM call to a JSON object."""

    def parse(self, text: str) -> Any:
        text = text.strip()
        try:
            return parse_json_markdown(text.strip(), parser=parse_partial_json)
        except json.JSONDecodeError as e:
            raise OutputParserError(f"Invalid json output: {text}") from e

    @property
    def _type(self) -> str:
        return "simple_json_output_parser"


class JsonListParser:
    """Parse the output of an LLM call to a list of JSON objects."""

    def parse(self, text: str) -> Any:
        text = text.strip()
        try:
            ret = parse_json_markdown(text.strip(), parser=parse_partial_json)
        except json.JSONDecodeError as e:
            raise OutputParserError(f"Invalid json listoutput: {text}") from e

        if not isinstance(ret, list):
            raise OutputParserError(f"Invalid json list output: {text}")

        return ret

    @property
    def _type(self) -> str:
        return "json_list_parser"

class SQLParser:
    """Parse the sql output of an LLM call.

    sql output is expected to be in the format:
    ---
    xxx
    ```sql
    <sql query>
    ```
    xxx
    ---

    Sometime the output is in the format:
    ---
    <sql query>
    ```
    xxxxxx
    ---
    """

    @property
    def _type(self) -> str:
        """Return the type key."""
        return "sql_parser"

    def parse(self, text: str) -> Dict[str, str]:
        """Parse the output of an LLM call.

        Args:
            text (str): output of an LLM call

        Returns:
            Dict[str, str]: a dict of parsed output
                - sql: the sql query
                - text: the original output text

        """
        origin_text = text
        sql_split =  text.split('```sql')
        if len(sql_split) > 1:
            text = sql_split[1]
        try:
            sql = text.split('```')[0]
        except IndexError:
            raise OutputParserError(f"Invalid sql output: {text}")
        return {"sql": sql, "text": origin_text}
